#!/usr/bin/env python
"""
File System Dump Generator for the Proof-of-Concept Implementation of a
"Preloading Read-Only File System"
(C) 2009 Martin J. Fiedler 
This software is public domain.
"""
import sys, os, zlib, fnmatch, stat, cPickle, cStringIO, profs

def fsobj_dump(obj, indent=""):
    """Produce a visual dump of an FSObject on the console."""
    name = obj.name
    extra = ""
    if   stat.S_ISDIR(obj.mode): name += "/"
    elif stat.S_ISLNK(obj.mode): name += "@ -> " + obj.__dict__.get('link', "???")
    elif stat.S_ISCHR(obj.mode): name += "-"
    elif stat.S_ISBLK(obj.mode): name += "+"
    elif stat.S_ISFIFO(obj.mode): name += "|"
    elif stat.S_ISSOCK(obj.mode): name += "="
    elif stat.S_ISREG(obj.mode): extra = " @ %d + %d = %d%%" % (obj.offset, obj.csize, not(obj.size) or (obj.csize * 100 / obj.size))
    print "%s%s (%d: %d%s)" % (indent, name, obj.inode, obj.size, extra)
    if stat.S_ISDIR(obj.mode) and ('objs' in obj.__dict__):
        return sum([fsobj_dump(sub, indent + "  ") for sub in obj.objs])
    elif stat.S_ISREG(obj.mode):
        return obj.size
    else:
        return 0

class mkprofs_error(RuntimeError):
    pass

class mkprofs(object):
    """ProFS filesystem creation class."""

    def __init__(self, dest, srcdir):
        """Initialize the instance and scan the source directory for files."""
        try:
            self.idx = open(dest + ".idx", "wb")
            self.dat = open(dest + ".dat", "wb")
        except IOError:
            print "Error: cannot open output files"
            raise mkprofs_error
        try:
            os.chdir(srcdir)
        except OSError:
            print "Error: cannot change into source directory"
            raise mkprofs_error
        self.inodes = 0
        self.offset = 0
        self.root = self.scan(".")
        if not self.root:
            raise mkprofs_error
        print self.countdirs(self.root), "objects total"

    def scan(self, path):
        """Scan a directory and add all files and subdirectories to the
        file system index."""
        try:
            s = os.lstat(path)
        except OSError, e:
            print "Error: cannot stat %s --" % path, e
            return None
        obj = profs.FSObject(os.path.basename(path.rstrip('/')))
        obj.mode = s.st_mode
        obj.uid = s.st_uid
        obj.gid = s.st_gid
        obj.atime = s.st_atime
        obj.mtime = s.st_mtime
        obj.ctime = s.st_ctime
        obj.rdev = s.st_rdev
        if obj.isfile() or obj.islink():
            obj.size = s.st_size
        if obj.islink():
            try:
                obj.link = os.readlink(path)
            except OSError, e:
                print "Error: cannot read target of symlink %s --" % path, e
                return None
        elif obj.isdir():
            print "scanning", "/" + path[2:], "..."
            try:
                files = os.listdir(path)
            except OSError, e:
                print "Error: cannot read contents of directory %s --" % path, e
                return None
            files.sort()
            obj.objs = filter(None, [self.scan(os.path.join(path, f)) for f in files])
        return obj

    def next(self):
        """Return the number of the next free inode."""
        self.inodes += 1
        return self.inodes

    def countdirs(self, obj):
        """Count the number of filesystem objects that are children of a
        specific FSObject. Or, in layman terms: Count all files and
        subdirectories in a directory."""
        if not obj.isdir():
            return 1
        obj.size = sum([self.countdirs(sub) for sub in obj.objs])
        return obj.size

    def add(self, path):
        """Adds a set of files to the data file."""
        self._add(self.root, ['*'] + path.lstrip('/').split('/'))

    def _add(self, obj, path, fullpath=""):
        """The actual implementation of the add() function."""
        if not(obj.isdir() or obj.isfile()):
            # we only handle files and directories here
            return
        if path:
            # path specified -> check if it matches and quit if it doesn't
            if not fnmatch.fnmatch(obj.name, path[0]):
                return
        if obj.isdir():
            # directory -> scan all subdirectories
            for sub in obj.objs:
                self._add(sub, path[1:], fullpath + "/" + sub.name)
            return
        if obj.inode:
            # already added -> skip
            return
        print "adding", fullpath, "..."
        try:
            f = open(fullpath[1:], "rb")
            data = zlib.compress(f.read(), 9)
            f.close()
        except IOError, e:
            print "Error reading %s --" % fullpath, e
            return
        try:
            self.dat.write(data)
        except IOError, e:
            print "Error adding %s --" % fullpath, e
            raise mkprofs_error
        obj.offset = self.offset
        obj.csize = len(data)
        obj.inode = self.next()
        self.offset += len(data)

    def add_inodes(self, obj):
        """Assign inode numbers to FSObjects that don't already have one
        (basically everything except normal files)."""
        if not obj.inode:
            obj.inode = self.next()
        if obj.isdir():
            for sub in obj.objs:
                self.add_inodes(sub)

    def finish(self):
        """Add the data of the remaining files, finalize and write the
        index file and close the data file."""
        self._add(self.root, ['*'])
        self.add_inodes(self.root)
        uncompressed = fsobj_dump(self.root)
        compressed = self.dat.tell()
        print "total size: uncompressed %s, compressed %s -> %d%%" % (uncompressed, compressed, compressed * 100 / uncompressed)
        print "writing index ..."
        idx = cStringIO.StringIO()
        cPickle.dump(self.root, idx, -1)
        csize = idx.tell()
        self.idx.write(zlib.compress(idx.getvalue()))
        print "index size: %d bytes (%d bytes uncompressed)" % (self.idx.tell(), csize)
        self.idx.close()
        self.dat.close()
        print "done."


if __name__ == "__main__":
    try:
        dest = sys.argv[1]
        srcdir = sys.argv[2]
    except IndexError:
        print "Usage:", sys.argv[0], "<dest_file_base> <source_directory> <priority_list>"
        print "<priority_list> is a text file with paths that shall be added to the pack first"
        sys.exit(2)

    # open and parse priority list file
    try:
        f = open(sys.argv[3], "r")
        prio = filter(None, [line.split('#', 1)[0].strip() for line in f.readlines()])
        f.close()
    except IndexError:
        prio = []
    except IOError, e:
        print "Error reading the priority list file:", e

    # create the filesystem
    try:
        m = mkprofs(dest, srcdir)
        for path in prio:
            m.add(path)
        m.finish()
    except mkprofs_error:
        sys.exit(1)
